﻿-- Load failure check
if not (Checklist2 and Checklist2.CORE_LOADED_SUCCESSFULLY) then return end

-- Checklist2 communications functions
local alreadyNagged = false;
local function nagforupdate(version)
  if version then
    version = " (|cffffff00"..version.."|r)"
  else
    version = ""
  end
  if not alreadyNagged then
    Checklist2:Print("A newer version of Checklist2 has been detected. "..version)
  end
end

local comm = {}
setmetatable(comm, Checklist2)

function Checklist2:OnCommReceive(prefix, sender, dist, versionID, vstring, cmd, ...)
  -- self:PrintLiteral(prefix, sender, dist, versionID, vstring, cmd, ...)
  if type(cmd) ~= "string" then return end
  -- UIErrorsFrame:AddMessage('|cffffff00Checklist2|r: Received data with cmd |cff00ff00"'..cmd..'"|r')
  -- self:PrintLiteral(versionID, self.db.profile.Settings.latestKnownVersionID)
  if versionID > Checklist2VersionCheck.latestKnownVerID then
    nagforupdate(vstring);
    Checklist2VersionCheck.latestKnownVerID = versionID;
    Checklist2VersionCheck.latestKnownVerString = vstring;
  end
  if comm[cmd] then
    comm[cmd](Checklist2, sender, ...)
  else
    self:Print(format("Received unknown command from \"%s\", who is using version %s. You may need to update.", sender, vstring))
  end
end



--[[
function comm:reqlists(sender, ...)
  self:Send("listdb", self.db.profile.Lists)
end
]]
local TEMPlist, TEMPname

StaticPopupDialogs["CHECKLIST2_ACCEPT_COMM_LIST_OVERWRITE"] = {
  text = "The list named \"%s\" received from %s will overwrite the list of the same name. Do you wish to continue?",
  button1 = ACCEPT,
  button2 = CANCEL,
  OnAccept = function()
    Checklist2.db.profile.Lists[TEMPname] = TEMPlist;
    Checklist2:ListItemSBUpdate()
    -- Checklist2:Print("Updated list |cffffff00"..TEMPname.."|r.")
    TEMPname, TEMPlist = nil, nil
  end,
  OnCancel = function()
    TEMPname, TEMPlist = nil, nil
  end,
  showAlert = 1,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1
};

function comm:listdb(sender, ...)
  for k,v in pairs(lists) do
    self.db.profile.Shared[k] = v;
  end
end

function comm:sendlist(sender, ...)
  local listName, list = ...
  -- self:PrintLiteral(listName, list,...)
  if not self.db.profile.Lists[listName] then
    self.db.profile.Lists[listName] = list;
    self:ListScrollBarUpdate()
    self:ListItemSBUpdate()
  else
    TEMPlist, TEMPname = list, listName
    StaticPopup_Show("CHECKLIST2_ACCEPT_COMM_LIST_OVERWRITE", listName, sender)
  end
end

function Checklist2:SendListToPlayer(player, listname)
  if type(player) ~= "string" or type(listname) ~= "string" then
    error("An error has occured while calling method \"Checklist2:SendListToPlayer()\"\n(incorrect argument types)")
    return
  end
  -- self:SendWhisper(player, 'sendlist', listname, self:GetList(listname))
  self:SendCommMessage("WHISPER", player, self.verID, self.verStr, 'sendlist', listname, self:GetList(listname));
end

function Checklist2:SendToDist(dist, cmd, ...)
  assert(type(dist) == "string" and type(cmd) == "string", "Checklist2:SendToDist() expects string args")
  self:SendCommMessage(dist, self.verID, self.verStr, cmd, ...)
end

function Checklist2:SendNote(dist, note, playerName)
  assert(type(dist) == "string", "Bad arg(s) to Checklist2:SendNote(dist, text, noteType, playerName)");
  assert(type(noteType) == "string", "Bad arg(s) to Checklist2:SendNote(dist, text, noteType, playerName)");
  assert(type(playerName) == "string", "Bad arg(s) to Checklist2:SendNote(dist, text, noteType, playerName)");
  
  local text, noteType
  local noteObject = self:GetNote(note);
  if not noteObject then
    return; -- The note does not exist.
  end
  text = noteObject.content;
  noteType = noteObject.type;
  
  if dist == "WHISPER" then
    self:SendCommMessage(dist, playerName, self.verID, self.verStr, 'sendnote', text, noteType)
    return;
  end
  
  self:SendCommMessage(dist, self.verID, self.verStr, 'sendnote', text, noteType)
end
