﻿Checklist2 = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceComm-2.0", "AceEvent-2.0")
Checklist2.name = "Checklist2"
--[[ Moved into their respective metatags in the toc file
Checklist2.credits = "Everyone who posted their suggestions on Curse and WoWI :D"
Checklist2.date = "Saturday February 9, 2008 8:54 AM GMT+8"
Checklist2.email = "quezpwns AT gmail DOT com"

--$Rev: 78724 $
--$Date: 2008-07-19 04:54:08 -0400 (Sat, 19 Jul 2008) $
--]]
local L = AceLibrary("AceLocale-2.2"):new("Checklist2")

Checklist2.version = "2.1.2.0830"
-- Used for the update checks performed on login (not yet written)
-- YYYYMMDDTTTT
-- Year, Month, Date, Time (24h format)
local INTERNAL_VERSION_ID   = 200804302004
local LISTBAR_ALWAYS_SHOWN  = true
local ITEMBAR_ALWAYS_SHOWN  = true
local ListInUse
local NoteInUse
local activeChecks
local activeItems
local initializeNotes   -- Notes initialization function

-- Time (in seconds) between checks if the current note is modified.
-- (for the "Save" button)
local NotesModifiedTextFrequency = 0.25

-- local aliases to globals
local getglobal = getglobal
local assert = assert
local format = format


Checklist2.versionID = INTERNAL_VERSION_ID;

-- Bindings
BINDING_HEADER_CHECKLIST2 = "Checklist2"
BINDING_NAME_SHOW_CHECKLIST2_UI = L["Toggle Checklist2 window"]

-- Ace2 slash command definitions
local options = {
    type = "group",
    desc = L["A checklist to keep track of tasks you need to do, now rewritten to include more features."],
    args = {
      ui = {
        name = "UI",
        type = 'execute',
        desc = L["Toggles the Checklist2 window"],
        func = function()
          Checklist2:ToggleUI()
        end,
        },
      reset = {
        name = L["Reset"],
        type = 'execute',
        desc = L["Resets the Checklist2 window's position, then shows it"],
        func = function()
          Checklist2_CoreFrame:ClearAllPoints()
          Checklist2_CoreFrame:SetPoint("CENTER", "UIParent", "CENTER")
          Checklist2SetupFrame:ClearAllPoints()
          Checklist2SetupFrame:SetPoint("CENTER", "Checklist2_CoreFrame", "CENTER")
          Checklist2_CoreFrame:Show()
          Checklist2:Print(L["Window positions have been reset."])
        end
        },
      newlist = {
        name = "New List",
        type = "text",
        desc = L["Creates a new empty list, ready for customization"],
        usage = "|cffffff00listname|r",
        get = false,
        set = function(listname)
          Checklist2:CreateList(listname)
          Checklist2:ListScrollBarUpdate()
        end
        },
      renamelist = {
        name = "Rename List",
        type = "text",
        desc = L["Change the name of the specified list"],
        usage = "|cffffff00old name, new name|r",
        get = false,
        set = function(val)
          local oldName, newName = strsplit(', ', val, 2)
          if (oldName and newName) then
            Checklist2:RenameList(oldName, newName)
          else
            self:Print(L["Unable to parse parameters to |cffffff00/checklist2 renamelist|r, please try again"])
          end
        end
      },
      deletelist = {
        name = "Delete List",
        type = "text",
        desc = L["Delete the specified list."],
        usage = "|cffffff00listname|r",
        get = false,
        set = function(listname)
          if listname == Checklist2:GetListInUse(true) then
            Checklist2:Print(format(L["Cannot delete the currently displayed list (|cffffff00%s|r)"], listname))
            return
          end
          if Checklist2:GetList(listname) then
            Checklist2.Delete = listname
            StaticPopup_Show("CHECKLIST2_DELETE_LIST", Checklist2.Delete)
          else
            Checklist2:Print("Could not find list \"|cffffff00" .. listname .. "|r\"")
          end
        end
        },
      setlist = {
        name = "setlist",
        type = "text",
        desc = L["Sets the list to display on the plugin tooltip and the main window."],
        usage = "|cffffff00listname|r",
        get = false,
        set = function(listname)
          Checklist2:SetListInUse(listname)
          Checklist2Fu:Update()
        end
      },
      check = {
        name = "Check",
        type = "text",
        desc = L["Checks the specified task number"],
        usage = L["|cffffff00tasknumber|r\n(only accepts numbers)"],
        get = false,
        set = function(taskID)
          taskID = tonumber(taskID)
          if not taskID then
            Checklist2:Print(L["Value given into |cffffff00/checklist2 check|r cannot be converted into a number."])
            return
          end
          taskID = math.floor(taskID)
          if taskID > 0 and taskID <= Checklist2:GetNumItems() then
            Checklist2:GetListInUse().checks["checked" .. taskID] = true
            Checklist2:ListItemSBUpdate()
            Checklist2Fu:Update()
          else
            Checklist2:Print(format(L["The specified task (|cffffff00%s|r) was not found."], taskID))
          end
        end
      },
      uncheck = {
        name = "Uncheck",
        type = "text",
        desc = L["Unchecks the specified task number"],
        usage = L["|cffffff00tasknumber|r\n(only accepts numbers)"],
        get = false,
        set = function(taskID)
          taskID = tonumber(taskID)
          if not taskID then
            Checklist2:Print(L["Value given into |cffffff00/checklist2 uncheck|r cannot be converted into a number."])
            return
          end
          taskID = math.floor(taskID)
          if taskID > 0 and taskID <= Checklist2:GetNumItems() then
            Checklist2:GetListInUse().checks["checked" .. taskID] = false
            Checklist2:ListItemSBUpdate()
            Checklist2Fu:Update()
          else
            Checklist2:Print(format(L["The specified task (|cffffff00%s|r) was not found."], taskID))
          end
        end
      }
    }
  }

Checklist2:RegisterChatCommand({"/checklist2", "/cl2"}, options)

function Checklist2:OnInitialize()
  -- Change the font size
  for i = 1, 15 do
    local itemtext = getglobal("Checklist2_CoreFrameContainer1CheckButton" .. i .. "Text")
    itemtext:SetFont("Fonts\\FRIZQT__.TTF", 12, " ")
    itemtext:SetWidth(645)
    itemtext:SetJustifyH("LEFT")
  end
  
  --self:Print("Checklist2 online!")
  
  if not Checklist2VersionCheck then
    Checklist2VersionCheck = {
      latestKnownVerID = INTERNAL_VERSION_ID,
      latestKnownVerString = self.version
    }
  end
  
  -- Version check
  -- self:PrintLiteral(Checklist2VersionCheck.latestKnownVerID, INTERNAL_VERSION_ID, Checklist2VersionCheck.latestKnownVerID > INTERNAL_VERSION_ID)
  if Checklist2VersionCheck.latestKnownVerID > INTERNAL_VERSION_ID then  -- Check if an update has been detected earlier
    local text
    if Checklist2VersionCheck.latestKnownVerString then
      text = "(|cff7fff7f" .. Checklist2VersionCheck.latestKnownVerString .. "|r)"
    else
      text = ""
    end
    self:Print(L["A newer version is available. "] .. text)
  end
  if Checklist2VersionCheck.latestKnownVerID < INTERNAL_VERSION_ID then  -- The user updated to a version higher than has been detected
    Checklist2VersionCheck.latestKnownVerID = INTERNAL_VERSION_ID
    Checklist2VersionCheck.latestKnownVerString = self.version
  end
  
  Checklist2_CoreFrameContainer1Texture1Texture:SetAlpha(.35)
  
  if self:GetNumLists() == 0 then
    self:CreateHelpList()
  end
  
  self:CollectListNames()
  
  -- Initialize the UI
  if type(self.db.profile.Settings.ListLastUsed) ~= "string" or not self:GetList(self.db.profile.Settings.ListLastUsed) then
    if #self.listNames > 0 then
      self.db.profile.Settings.ListLastUsed = self.listNames[1];
    end
  end
  self:SetListInUse(self.db.profile.Settings.ListLastUsed)
  self:SetCommPrefix("Checklist2")
  if not self.db.profile.Settings.versionID then
    self.db.profile.Settings.versionID = 0;
  end
  self:UpdateTasksIndicator()
  for i = 1, 30 do
    local button = getglobal("Checklist2AppletListContainerButton" .. i)
    if button.list == ListInUse then
      button:LockHighlight();
      break;
    end
  end
  initializeNotes(self)
  
  local tabID = self.db.profile.Settings.tabDisplayed;
  if type(tabID) == "number" then
    PanelTemplates_SetTab(Checklist2_CoreFrame, tabID);
    Checklist2_CoreFrameContainer1:Hide();
    Checklist2_CoreFrameContainer2:Hide();
    getglobal("Checklist2_CoreFrameContainer" .. tabID):Show();
  end
  
  self.verID = Checklist2VersionCheck.latestKnownVerID;
  self.verStr = Checklist2VersionCheck.latestKnownVerString;
end

function Checklist2:OnEnable()
  self:RegisterComm("Checklist2", "GROUP")
  -- "GROUP" distribution chooses between BATTLEGROUND, PARTY and RAID, depending
  -- on which is most applicable.
  -- Received messages are passed into method OnCommReceive()
  self:RegisterComm("Checklist2", "WHISPER")
  self:RegisterComm("Checklist2", "GUILD")
end

function Checklist2:OnProfileEnable()
  -- Update the UI
  if self:GetNumLists() == 0 then
    self:CreateHelpList()
  end
  if self:GetNumNotes() == 0 then
    self:CreateHelpNote();
  end
  self:SetListInUse(self.db.profile.Settings.ListLastUsed)
  self:SetNoteInUse(self.db.profile.Settings.noteDisplayed)
  self:CollectListNames()
  self:ListScrollBarUpdate()
  self:ListItemSBUpdate()
end

-- Checklist2 core functions
function Checklist2:ToggleUI()
  if Checklist2SetupFrame:IsVisible() then
    return
  end
  if not Checklist2_CoreFrame:IsVisible() then
    Checklist2_CoreFrame:Show()
    UIFrameFadeIn(Checklist2_CoreFrame, 0.075, 0, 1)
  else
    Checklist2_CoreFrame:Hide()
  end
end



-- Revamped SetListInUse method
-- List names are now used as identifiers instead of a numeric ID
function Checklist2:SetListInUse(listName, override)
  if listName == ListInUse and not override then
    return  -- Save ourselves the effort
  end
  -- ChatFrame1:AddMessage(tostring(listName))
  if Checklist2SetupFrame:IsVisible() then
    --self:Print("Cannot change lists while the setup dialog is open.")
    return false
  end
  
  assert(type(listName) == "string", "Invalid argument passed into Checklist2 method SetListInUse (expecting string, got " .. type(listName) .. ")")
  
  if not self:GetList(listName) then  -- Safe failure
    self:Print(format(L["The list \"%s\" was not found."], listName))
    return
  end
  activeChecks = self:GetList(listName).checks
  activeItems = self:GetList(listName).items
  ListInUse = listName
  self:ListItemSBUpdate()
  self.db.profile.Settings.ListLastUsed = listName
  Checklist2_CoreFrameContainer1Label1Label:Show()
  Checklist2_CoreFrameContainer1Label1Label:SetText(listName .. " - Checklist v"..self.version)
  self:ListScrollBarUpdate()
  self:UpdateTasksIndicator()
end

function Checklist2:GetNumItems()
  local numItems = 0
  for k,v in pairs(self:GetListInUse().items) do
    numItems = numItems + 1
  end
  return numItems
end


function Checklist2:GetNumLists()
  local lists = 0
  local liststbl = self.db.profile.Lists
  for k,v in pairs(liststbl) do
    lists = lists + 1
  end
  return lists
end

function Checklist2:GetListInUse(giveName)
  if giveName then
    return ListInUse
  end
  return self.db.profile.Lists[ListInUse]
end

function Checklist2:GetList(list)
  -- self:PrintLiteral(list)
  return self.db.profile.Lists[list]
end


function Checklist2:CheckAll()
  for i = 1, self:GetNumItems() do
    self:CheckClicked(i, true, true)
  end
  self:UpdateTasksIndicator()
  self:ListItemSBUpdate()
  Checklist2Fu:Update()
end

function Checklist2:UncheckAll()
  for i = 1, self:GetNumItems() do
    self:CheckClicked(i, false, true)
  end
  self:UpdateTasksIndicator()
  self:ListItemSBUpdate()
  Checklist2Fu:Update()
end

function Checklist2:CheckClicked(checkID, isChecked, dontUpdateUI)
  -- convert nil to false, 1 to true
  -- workaround for a bug where only a portion of a list gets modified when
  -- "Check All" or "Uncheck All" gets clicked
  -- PS: Double negatives should not be used in English :)
  self:GetListInUse().checks["checked" .. tostring(checkID)] = not (not isChecked)
  if not dontUpdateUI then
    self:UpdateTasksIndicator()
  end
end

function Checklist2:GetPercentageChecked()
  local numTasks, numChecked = 0, 0
  for k,v in pairs(self:GetListInUse().checks) do
    numTasks = numTasks + 1
    if v then
      numChecked = numChecked + 1
    end
  end
  local percChecked = (numChecked / numTasks) * 100
  percChecked = (floor(percChecked*10)/10)
  percChecked = percChecked .. "%"
  return percChecked
end

function Checklist2:UpdateTasksIndicator()
  -- Really freakin' long! XD
  Checklist2_CoreFrameContainer1AddRemoveItemsButtonFS:SetText(format(L["Tasks: %i/%i (%s)"], self:GetNumTasksChecked(), self:GetNumItems(), self:GetPercentageChecked()))
end

-- List menu GUI update function
function Checklist2:ListScrollBarUpdate()
  if not self.listNames then
    self:CollectListNames()
  end
  local line;
  local lineplusoffset;
  local names;
  local text;
  FauxScrollFrame_Update(Checklist2AppletScrollBar, Checklist2:GetNumLists(), 30, 16, nil, nil, nil, nil, nil, nil, LISTBAR_ALWAYS_SHOWN);
  for line = 1,30 do
    local button = getglobal("Checklist2AppletListContainerButton" .. line);
    lineplusoffset = line + FauxScrollFrame_GetOffset(Checklist2AppletScrollBar);
    if lineplusoffset <= Checklist2:GetNumLists() then
      text = self.listNames[lineplusoffset]
      button:SetText(text);
      button.list = text;
      button:Show();
      if button.list ~= ListInUse then
        button:UnlockHighlight();
      else
        button:LockHighlight();
      end
    else
      button:Hide();
    end
  end
end

-- List item GUI updater function
function Checklist2:ListItemSBUpdate()
  local numberOfItems = self:GetNumItems();
  local line;
  local lineplusoffset;
  FauxScrollFrame_Update(Checklist2ChecksScrollBar, numberOfItems, 15, 32, nil, nil, nil, nil, nil, nil, ITEMBAR_ALWAYS_SHOWN);
  for line=1,15 do
    local listItem = getglobal("Checklist2_CoreFrameContainer1CheckButton" .. tostring(line));
    local listItemText = getglobal(listItem:GetName() .. "Text");
    local setButton = getglobal("Checklist2SetButton" .. line);
    lineplusoffset = line + FauxScrollFrame_GetOffset(Checklist2ChecksScrollBar);
    if lineplusoffset <= numberOfItems then
      setButton:SetText(tostring(lineplusoffset) .. ".");
      setButton:Show();
      setButton:SetID(lineplusoffset);
      listItemText:SetText(activeItems["check" .. lineplusoffset .. "Text"]);
      listItem:SetChecked(self:GetListInUse().checks["checked" .. lineplusoffset]);
      listItem:SetID(lineplusoffset);
      listItem:Show();
    else
      setButton:Hide();
      listItem:Hide();
    end
  end
end

function Checklist2:CreateList(listName, norand)
  assert(type(listName) == "string", "Invalid argument passed into Checklist2 method \"CreateList\" (expecting string, got " .. type(listName) .. ")")
  if not norand then
    self.db.profile.Lists[listName] = self:GenerateRandomList()
  else
    self.db.profile.Lists[listName] = {
      checks = { checked1   = false },
      items  = { check1Text = self.Sayings[random(1, #self.Sayings)] }
    }
  end
  self:CollectListNames()
  self:ListScrollBarUpdate()
end

function Checklist2:CreateListItem(dontUpdateGUI)
  local x = self:GetNumItems() + 1
  self:GetListInUse().checks["checked" .. x] = false
  self:GetListInUse().items["check" .. x .. "Text"] = self.Sayings[random(1, #self.Sayings)]
  if dontUpdateGUI then
    -- For create list items "en masse"
    -- Updating the GUI is slow, so it's best to skip it when this
    -- function will get called repeatedly.
    return
  end
  self:ListItemSBUpdate()
  Checklist2Fu:Update()
  self:UpdateTasksIndicator()
end

function Checklist2:ReduceListItems(dontUpdateGUI)
  local x = self:GetNumItems()
  if x > 1 then
    self:GetListInUse().checks["checked" .. x] = nil
    self:GetListInUse().items["check" .. x .. "Text"] = nil
    if not dontUpdateGUI then
      self:ListItemSBUpdate()
      Checklist2Fu:Update()
      self:UpdateTasksIndicator()
    end
  else
    self:Print(L["There is only one task left!"])
  end
end

function Checklist2:ReduceListItemsEnMasse(num)
  local x = self:GetNumItems()
  num = tonumber(num)
  if not num then
    self:Print(L["Entered value cannot be converted into a number"])
    return
  end
  if num > 50 or num > (Checklist2:GetNumItems() - 1) then
    Checklist2:Print(L["Cannot remove more than 50 tasks, or the amount of tasks currently existing."])
    return
  end
  for i = 1, num do -- this shouldn't execute if it's not a number. Sanity checks above.
    self:GetListInUse().checks["checked" .. tostring(x)] = nil
    self:GetListInUse().items["check" .. tostring(x) .. "Text"] = nil
    x = x - 1
  end
  self:ListItemSBUpdate()
  Checklist2Fu:Update()
  self:UpdateTasksIndicator()
end

function Checklist2:OpenSetup(id)
  Checklist2EB_ConfirmText:SetFormattedText(L["You are editing options for task %s in list \"%s\".\nEnter text below, or press Escape to exit."], id, ListInUse)
  Checklist2.OPEN_ITEM_ID = id
  Checklist2SetupFrameEB:SetText(Checklist2:GetListInUse().items["check" .. Checklist2.OPEN_ITEM_ID .. "Text"])
  Checklist2SetupFrame:Show()
end

--[[
function Checklist2:CloseSetupFrame()
  
end

function Checklist2:HideWindow()
  Checklist2_CoreFrame:Hide()
  Checklist2_CoreFrame:SetAlpha(1)
end

function Checklist2:ShowWindow()
  Checklist2_CoreFrame:Show()
end
--]]

function Checklist2:RenameList(oldName, newName)
  assert((type(oldName == "string") and type(newName) == "string"), "Bad argument(s) to method Checklist2:RenameList(oldName, newName)")
  
  if newName == "" then
    -- self:Print("Lists cannot be renamed to \"\".")
    StaticPopup_Show("CHECKLIST2_CANNOT_RENAME_TO_EMPTY_STRING")
    return
  end
  
  if not self:GetList(oldName) then
    self:Print(L["The specified list was not found."])
    return
  end
  
  if not self:GetList(newName) then
    if oldName == ListInUse then
      Checklist2_CoreFrameContainer1Label1Label:SetText(newName .. " - Checklist v" .. self.version);
      ListInUse = newName;
      self.db.profile.Settings.ListLastUsed = ListInUse; -- Fixes list not loading if renamed then reloading the UI/relogging
    end
    self.db.profile.Lists[newName] = self:GetList(oldName)
    self.db.profile.Lists[oldName] = nil
    self:CollectListNames()
    self:ListScrollBarUpdate()
  else
    self:Print(format(L["A list named \"%s\" already exists."], newName))
  end
  
  Checklist2Fu:Update()
end

function Checklist2:DeleteList(list)
  if type(list) ~= "string" then
    error("Checklist2:DeleteList(): Expecting string, got " .. type(list))
  end
  if list ~= ListInUse then
    self.db.profile.Lists[list] = nil
    self:CollectListNames()
    self:ListScrollBarUpdate()
  else
    self:Print(format(L["Cannot delete the currently displayed list (|cffffff00%s|r)"], ListInUse))
  end
end

function Checklist2:Switch(item1, item2)
  if not self:GetListInUse().items["check" .. item2 .. "Text"] then
    self:Print(L["That task is already at either the top or the bottom of the list."])
    return
  end
  local check, item = self:GetListInUse().checks["checked" .. item2], self:GetListInUse().items["check" .. item2 .. "Text"]
  --[[ debug stuff
  self:PrintLiteral(check, item)
  self:PrintLiteral(self:GetListInUse().checks["checked" .. item1], self:GetListInUse().items["check" .. item1 .. "Text"])
  --]]
  self:GetListInUse().checks["checked" .. item2] = self:GetListInUse().checks["checked" .. item1]
  self:GetListInUse().items["check" .. item2 .. "Text"]  = self:GetListInUse().items["check" .. item1 .. "Text"]
  self:GetListInUse().checks["checked" .. item1] = check
  self:GetListInUse().items["check" .. item1 .. "Text"]  = item
  if Checklist2SetupFrame:IsVisible() then
    self:ListItemSBUpdate()
    Checklist2.OPEN_ITEM_ID = item2
    Checklist2EB_ConfirmText:SetFormattedText(L["You are editing options for task %s in list \"%s\".\nEnter text below, or press Escape to exit."], item2, ListInUse)
  end
  Checklist2Fu:Update()
end

StaticPopupDialogs["CHECKLIST2_DELETE_LIST"] = {
  text = L["Are you sure you want to delete the list \"%s\"?"],
  button1 = ACCEPT,
  button2 = CANCEL,
  OnAccept = function()
      Checklist2:DeleteList(Checklist2.Delete); 
      -- .Delete is set by clicking on the "Delete" dropdown button.
  end,
  timeout = 0,
  showAlert = 1,
  whileDead = 1,
  hideOnEscape = 1
};

function Checklist2:CollectListNames()
  if not self.listNames then
    self.listNames = {}
  end
  -- This manually clears the table, allowing Checklist2 to recycle it.
  -- This method of emptying a table takes more and more time as the table
  -- gets larger.
  for i,v in ipairs(self.listNames) do
    self.listNames[i] = nil;
  end
  for k,v in pairs(self.db.profile.Lists) do
    if (type(v) == "table") then -- Check if it hasn't been deleted or tampered with.
      -- tinsert(self.listNames, k) -- Insert the table key, which is the name of the list
      self.listNames[#self.listNames + 1] = k
    end
  end
  table.sort(self.listNames) -- Alphabetize
end

function Checklist2:GetNumTasksChecked(list)
  -- self:Print(debugstack())
  local checks = self:GetListInUse().checks
  local numChecked = 0
  for k,v in pairs(checks) do
    if v then
      numChecked = numChecked + 1
    end
  end
  return numChecked
end

-- Debug function that shows the 30 list item container buttons.
-- Primarily used for positioning.

--[[
function cl2showAll()
  for line = 1,30 do
    getglobal("Checklist2AppletListContainerButton" .. line):SetText("List Container Button " .. line);
    getglobal("Checklist2AppletListContainerButton" .. line):Show();
  end
end
--]]

-- Helper function
-- http://www.wowwiki.com/Tremovebyval
--[[
function Checklist2:tremovebyval(tab, val)
  for k,v in tab do
    if(v==val) then
      table.remove(tab, k);
      return true;
    end
  end
  return false;
end
--]]


function Checklist2:TabHandler(id)
  if not type(id) == "number" then 
    error("Bad arg #2 passed into :TabHandler (expecting number, got " .. type(id) .. ")")
  end
  PlaySound("igCharacterInfoTab")
  PanelTemplates_SetTab(Checklist2_CoreFrame, id)
  self["OnTab" .. id .. "Click"](self)
  self.db.profile.Settings.tabDisplayed = id
end

function Checklist2:OnTab1Click()
  Checklist2_CoreFrameContainer1:Show()
  Checklist2_CoreFrameContainer2:Hide()
end

function Checklist2:OnTab2Click()
  Checklist2_CoreFrameContainer1:Hide()
  Checklist2_CoreFrameContainer2:Show()
end

--]]

-- =================================================================
--  Notes code
-- =================================================================

function initializeNotes(self)
  self:CollectNoteNames()
  self.notesEditBox = Checklist2_CoreFrameContainer2NotesEBScrollBarEditBox
  if self:GetNumNotes() == 0 then
    self:CreateHelpNote();
  end
  self:SetNoteInUse(self.db.profile.Settings.noteDisplayed)
  self.notesInitialized = true
end

function Checklist2:CollectNoteNames()
if not self.noteNames then
    self.noteNames = {};
  end
  for i,v in pairs(self.noteNames) do
    self.noteNames[i] = nil;
  end
  for k,v in pairs(self.db.profile.Notes) do
    if type(v) == "table" then
      self.noteNames[#self.noteNames + 1] = k;
    end
  end
end

function Checklist2:GetNumNotes()
  local notes = 0
  local notestbl = self.db.profile.Notes
  for k,v in pairs(notestbl) do
    notes = notes + 1
  end
  return notes
end

StaticPopupDialogs["CHECKLIST2_NOTES_CONFIRM_LIST_CHANGE_WITHOUT_SAVE"] = {
  text = "Are you sure you want to change lists without saving?",
  button1 = ACCEPT,
  button3 = "Save and continue",
  button2 = CANCEL,
  OnAccept = function()
    Checklist2:SetNoteInUse(Checklist2.destNote, true)
  end,
  OnAlt = function()
    Checklist2:SaveNote();
    Checklist2:SetNoteInUse(Checklist2.destNote)
  end,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1
};
StaticPopupDialogs["CHECKLIST2_NOTES_CONFIRM_REVERT"] = {
  text = "Are you sure you want to revert your unsaved edits to this note?",
  button1 = ACCEPT,
  button2 = CANCEL,
  OnAccept = function()
    Checklist2.notesEditBox:SetText(Checklist2:GetNoteInUse().content);
    Checklist2.notesEditBox:ClearFocus();
  end,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1
}

function Checklist2:SetNoteInUse(note, overrideSaveCheck)
  assert(type(note) == "string", "Checklist2:SetNoteInUse(list) - 'list' must be a string");
  if note == NoteInUse then
    return;
  end
  if self.notesInitialized then
    if (self.notesEditBox:GetText() ~= self:GetNote(NoteInUse, true) and not overrideSaveCheck) then
      self.destNote = note;
      StaticPopup_Show("CHECKLIST2_NOTES_CONFIRM_LIST_CHANGE_WITHOUT_SAVE");
      return;
    end
  end
  NoteInUse = note;
  Checklist2_CoreFrameContainer2TitleText:SetText(NoteInUse  ..  " - Checklist Notes")
  self:SetNotesMode(self:GetNote(NoteInUse).type)
  self.db.profile.Settings.noteDisplayed = note;
  self.notesEditBox:SetText(self.db.profile.Notes[note].content);
  self:NotesSelectScrollBarUpdate();
  self.notesEditBox:ClearFocus();
  -- self.notesEBSB:SetVerticalScroll(0);
  self.forceNotesToTop = 0;
end

function Checklist2:CheckIfNoteModified(frame, elapsed)
  if not frame or self.pauseModifiedNotesCheck then
    return
  end
  frame.TimeSinceLastUpdate = frame.TimeSinceLastUpdate + elapsed; 	
  
  if self.forceNotesToTop < 3 then
    self.notesEBSB:SetVerticalScroll(0);
    self.forceNotesToTop = self.forceNotesToTop + 1
  end
  
  if (frame.TimeSinceLastUpdate > NotesModifiedTextFrequency) then
    frame.TimeSinceLastUpdate = 0;
    
    if self.notesEdit:GetText() ~= self:GetNote(NoteInUse, true) then
      Checklist2_CoreFrameContainer2SaveNoteButton:Enable();
      Checklist2_CoreFrameContainer2RevertEditsButton:Enable();
    else
      Checklist2_CoreFrameContainer2SaveNoteButton:Disable();
      Checklist2_CoreFrameContainer2RevertEditsButton:Disable();
    end
  end
end

function Checklist2:GetNoteInUse(getName)
  -- self:Print(debugstack())
  if getName then
    return NoteInUse
  else
    return self.db.profile.Notes[NoteInUse]
  end
end

function Checklist2:GetNote(note, getText)
  -- self:Print(debugstack())
  assert(type(note) == "string", "Usage: Checklist2:GetNote(note, getText) - \"note\" must be a string, getText can be true, false, or nil/ommited.")
  if not getText then
    return self.db.profile.Notes[note]
  else
    return self.db.profile.Notes[note].content
  end
end

function Checklist2:NotesSelectScrollBarUpdate()
  if not self.noteNames then
    self:CollectNoteNames()
  end
  local line;
  local lineplusoffset;
  local names;
  local text;
  FauxScrollFrame_Update(Checklist2_CoreFrameContainer2NotesScrollBar, Checklist2:GetNumNotes(), 30, 16, nil, nil, nil, nil, nil, nil, LISTBAR_ALWAYS_SHOWN);
  for line = 1,30 do
    local button = getglobal("Checklist2_CoreFrameContainer2NotesContainerButton" .. line);
    lineplusoffset = line + FauxScrollFrame_GetOffset(Checklist2_CoreFrameContainer2NotesScrollBar);
    if lineplusoffset <= Checklist2:GetNumNotes() then
      text = self.noteNames[lineplusoffset]
      button:SetText(text);
      button.note = text;
      button:Show();
      if button.note ~= self.db.profile.Settings.noteDisplayed then
        button:UnlockHighlight();
      else
        button:LockHighlight();
      end
    else
      button:Hide();
    end
  end
end

function Checklist2:SaveNote()
  -- Notes work differently compared to lists. List changes are saved 
  -- the moment they are performed (changing task text, task status, 
  -- task ordering, list name, etc)
  -- We do not require any extra args as we only need to save the
  -- current note.
  self.db.profile.Notes[NoteInUse].content = Checklist2_CoreFrameContainer2NotesEBScrollBarEditBox:GetText();
end

function Checklist2:SetNotesMode(mode)
  -- self:Print(debugstack())
  assert(type(mode) == "number", "Usage: Checklist2:SetNotesMode(mode) - \"mode\" must be an integer. 1 is for text mode, 2 is for Lua mode.");
  self.notesMode = mode;
  if mode == 1 then
    Checklist2_CoreFrameContainer2ModeSelectButtonText:SetText("Text Mode");
    self:GetNote(NoteInUse).type = mode;
    Checklist2_CoreFrameContainer2RunScriptButton:Disable();
    Checklist2_CoreFrameContainer2ImportListButton:Disable();
  else
    Checklist2_CoreFrameContainer2ModeSelectButtonText:SetText("Lua Mode")
    self:GetNote(NoteInUse).type = mode;
    Checklist2_CoreFrameContainer2RunScriptButton:Enable();
    Checklist2_CoreFrameContainer2ImportListButton:Enable();
  end
end

function Checklist2:RenameNote(oldName, newName)
  assert((type(oldName == "string") and type(newName) == "string"), "Bad argument(s) to method Checklist2:RenameNote(oldName, newName)")
  
  if newName == "" then
    StaticPopup_Show("CHECKLIST2_CANNOT_RENAME_TO_EMPTY_STRING");
    return;
  end
  
  if not self:GetNote(oldName) then
    self:Print("The specified note was not found.")
  end
  if not self:GetNote(newName) then
    if oldName == NoteInUse then
      Checklist2_CoreFrameContainer2TitleText:SetText(newName .. " - Checklist v" .. self.version);
      NoteInUse = newName;
      self.db.profile.Settings.noteDisplayed = newName;
    end
    self.db.profile.Notes[newName] = self.db.profile.Notes[oldName]
    self.db.profile.Notes[oldName] = nil
    self:CollectNoteNames()
    self:NotesSelectScrollBarUpdate()
  else
    self:Print(format("A note named \"%s\" already exists.", newName))
  end
end

function Checklist2:CreateNote(name)
  if name == "" then return; end
  self.db.profile.Notes[name] = {
            content = "",
            type = 1
            }
  self:CollectNoteNames();
  self:NotesSelectScrollBarUpdate();
  return self.db.profile.Notes[name]
end


function Checklist2:DeleteNote(note)
  self.db.profile.Notes[note] = nil;
  self:CollectNoteNames();
  self:NotesSelectScrollBarUpdate();
end

StaticPopupDialogs["CHECKLIST2_CANNOT_RENAME_TO_EMPTY_STRING"] = {
  text = "You cannot rename a list or note to \"\"",
  button1 = OKAY,
  timeout = 0,
  showAlert = 1,
  whileDead = 1,
  hideOnEscape = 1
};

StaticPopupDialogs["CHECKLIST2_CANNOT_DELETE_CURRENT"] = {
  text = "You cannot delete the list or note currently being displayed.",
  button1 = OKAY,
  timeout = 0,
  showAlert = 1,
  whileDead = 1,
  hideOnEscape = 1
};

StaticPopupDialogs["CHECKLIST2_LIST_IMPORT_SYNTAX_ERROR"] = {
  text = "A syntax error has been found in the list's code:\n\n%s",
  button1 = OKAY,
  timeout = 0,
  showAlert = 1,
  whileDead = 1,
  hideOnEscape = 1
};

StaticPopupDialogs["CHECKLIST2_GENERIC_MESSAGE"] = {
  text = "%s",
  button1 = OKAY,
  timeout = 0,
  showAlert = 1,
  whileDead = 1,
  hideOnEscape = 1
};

StaticPopupDialogs["CHECKLIST2_DELETE_NOTE"] = {
  text = "Are you sure you want to delete the note \"%s\"?",
  button1 = ACCEPT,
  button2 = CANCEL,
  OnAccept = function()
      Checklist2:DeleteNote(Checklist2.deleteNote); 
      -- .deleteNote is set by clicking on the "Delete" dropdown button.
  end,
  timeout = 0,
  showAlert = 1,
  whileDead = 1,
  hideOnEscape = 1
};

StaticPopupDialogs["CHECKLIST2_LIST_IMPORT_NAME_PROMPT"] = {
  text = "What would you like to name this list?",
  button1 = ACCEPT,
  button2 = CANCEL,
  timeout = 0,
  -- showAlert = 1,
  whileDead = 1,
  hideOnEscape = 1,
  hasEditBox = 1,
  EditBoxOnEscapePressed = function()
    this:GetParent():Hide();
  end,
  EditBoxOnEnterPressed = function()
    local listName = getglobal(this:GetParent():GetName() .. "EditBox"):GetText()
    if listName == "" then
      Checklist2:Message("List name cannot be \"\".")
      StaticPopup_Show("CHECKLIST2_LIST_IMPORT_NAME_PROMPT")
      return
    end
    Checklist2:ImportList(listName)
    this:GetParent():Hide();
  end,
  OnAccept = function()
    local listName = getglobal(this:GetParent():GetName() .. "EditBox"):GetText()
    if listName == "" then
      Checklist2:Message("List name cannot be \"\".")
      StaticPopup_Show("CHECKLIST2_LIST_IMPORT_NAME_PROMPT")
      return
    end
    Checklist2:ImportList(listName)
  end
};

StaticPopupDialogs["CHECKLIST2_VIEW_IMPORTED_LIST"] = {
  text = "The list was successfully imported. Would you like to view it now?",
  button1 = ACCEPT,
  button2 = CANCEL,
  OnAccept = function()
    PanelTemplates_SetTab(Checklist2_CoreFrame, 1);
    Checklist2_CoreFrameContainer1:Show();
    Checklist2_CoreFrameContainer2:Hide();
    PlaySound("igCharacterInfoTab");
    Checklist2:SetListInUse(Checklist2.importedList)
  end,
  timeout = 0,
  showAlert = 1,
  whileDead = 1,
  hideOnEscape = 1
};

function Checklist2:Message(str)
  StaticPopup_Show("CHECKLIST2_GENERIC_MESSAGE", tostring(str))
end

-- ================================================
-- List import/export functions
-- ================================================

function Checklist2:ExportList(list)
  assert(type(list) == "string", "Invalid arg to Checklist2:ExportList(list) (string expected)")
  listObj = self:GetList(list);
  
  assert(listObj, "List does not exist");
  
  -- Prepare
  local liststr = ("{\n  checks = {")
  local numTasks = self:GetNumItems();
  local tObj, csObj = listObj.items, listObj.checks     -- locals stand for "Tasks Object" and "Check Status Object"
  
  -- Generate task status data
  for i = 1, numTasks do
    liststr = liststr  ..  ("\n   checked"  ..  i  ..  " = "  ..  tostring( csObj["checked"  ..  i] )  ..  ",")
  end
  
  liststr = liststr  ..  "\n  },\n  items = {"
  
  -- Generate text data
  local text;
  for i = 1, numTasks do
    text = tostring( tObj["check"  ..  i  ..  "Text"] )
    assert(type(text) == "string", "Encountered an error while generating list data (task text is not a string, list may be corrupted)")
    
    -- Necessary substitutions
    text = text:gsub("\\", "\\\\");
    text = text:gsub("\"", "\\\"");
    text = text:gsub("\124", "\\124");
    text = text:gsub("\n", "\\n");
    text = text:gsub("\010", "\\010");
    
    liststr = liststr  ..  ("\n    check"  ..  i  ..  "Text = \""  ..  text  ..  "\",");
  end
  
  liststr = liststr  ..  "\n  }\n}";
  collectgarbage("collect"); -- We generated a LOT of garbage, increases with list size.
  
  local note = self:CreateNote("Exported \"" .. list .. "\"");
  note.content = liststr;
  
  -- Change tabs and select our new exported list
  PanelTemplates_SetTab(Checklist2_CoreFrame, 2);
  Checklist2_CoreFrameContainer1:Hide();
  Checklist2_CoreFrameContainer2:Show();
  PlaySound("igCharacterInfoTab")
  self:SetNoteInUse("Exported \"" .. list .. "\"")
end

function Checklist2:ImportList(listName)
  assert(type(listName), "Bad argument to :ImportList(listName) - listName must be a string")
  -- self:PrintLiteral(listName)
  
  -- Necessary substitutions
  local text = listName
  text = text:gsub("\\", "\\\\");
  text = text:gsub("\"", "\\\"");
  -- text = text:gsub("\124", "\\124");
  text = text:gsub("\n", "\\n");
  text = text:gsub("\010", "\\010");
  
  local listData = ("Checklist2.db.profile.Lists[\"" .. text .. "\"] = " .. self:GetNote(NoteInUse, true))
  -- self:PrintLiteral(listData)
  -- self:Print(listData)
  -- self:Print(debugstack())
  local func, errormsg = loadstring(listData, listName)
  -- self:PrintLiteral(func, errormsg)
  -- self:Print(debugstack())
  if (not func) and (errormsg or error("func and errormsg is nil! Whiskey!\nTango!\nFoxtrot!")) then
    -- Above error message should never fire unless the loadsting function is glitched.
    -- If it does, it should offer some laughs :)
    StaticPopup_Show("CHECKLIST2_LIST_IMPORT_SYNTAX_ERROR", errormsg);
  else
    if self:GetList(listName) then
      self:Message("There is already another list with that name.");
      return;
    end
    func();
    self:CollectListNames();
    self:ListScrollBarUpdate();
  end
  self.importedList = listName;
  StaticPopup_Show("CHECKLIST2_VIEW_IMPORTED_LIST");
end

-- "Leave a message" to the other files signaling that the core loaded succesfully
Checklist2.CORE_LOADED_SUCCESSFULLY = true;

-- Debug purposes only!
-- FrameXML_Debug(1)